/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: bulk-import.c,v 1.36 2005/11/16 23:29:40 theno23 Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <mysql.h>
#include <mysqld_error.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <raptor.h>
#include <getopt.h>
#include <glib.h>
#include <db.h>
/* for fink+osx you will need, should autoconf that
#include <db4/db.h>
*/

#include "config.h"
#include "tstore.h"
#include "util.h"
#include "inference.h"
#include "bulk-import.h"
#include "timing.h"
#include "rdf_constants.h"

#define MEMBER_PREFIX "http://www.w3.org/1999/02/22-rdf-syntax-ns#_"
#define MEMBER_PROP   "http://www.w3.org/2000/01/rdf-schema#member"

#define INTEGER_DT    "http://www.w3.org/2001/XMLSchema#integer"

#if MYSQL_VERSION_ID >= 40100
#define SQL_TRUE  "true"
#define SQL_FALSE "false"
#else
#define SQL_TRUE  "1"
#define SQL_FALSE "0"
#endif

#define TMP_SIZE 256

#define CLASS_H_SIZE 1024       /* size of the running hash of classes that
				 * have been checked to see if thier declared */

/* indexes into various arrays for the three consistency indexes */
#define T 0
#define S 1
#define M 2

#ifdef ENABLE_QUADS
#define TB 1
#else
#define TB 0
#endif

#define CACHE_FILE_TEMPLATE "/tmp/%s_%c%d.bdb"
#define SUBBLOCK_FILE_TEMPLATE "/tmp/%s_%c_sb%d.dat"
#define DAT_FILE_TEMPLATE "/tmp/%s_%c.dat"
#define LIVE_DAT_FILE_TEMPLATE "/tmp/%s_%c_live.dat"

static const char *cache_letters = "tsm";

typedef struct {
    int verbosity;
    int count_err;
    int count_warn;
    int count_trip;
    int tax_rebuild;
    ts_hash model;
    ts_connection *db;
    FILE *msg;
    ts_hash checked_classes[CLASS_H_SIZE];
} ts_bulk_parse_stuff;

static void store_stmt(void *user_data, const raptor_statement *statement);

static void fatal_err_handler(void *data, raptor_locator *locator,
                          const char *message);
static void err_handler(void *data, raptor_locator *locator,
                          const char *message);

static void warning_handler(void *data, raptor_locator *locator,
                            const char *message);

static ts_id check_symbol(ts_connection *c, ts_hash hash, char *lex,
			    int dt, int lang);

static ts_id check_symbol_later(ts_connection *c, ts_hash hash, char *lex,
			    int dt, int lang);

static void ts_bulk_assert_triple(ts_connection *c, ts_bulk_parse_stuff *data,
				  char *subj, char *pred, char *obj, 
				  ts_obj_type type, int dt, int lang);

static void parse_subblock(ts_connection *c, int block);

static void wait_load();
static void schedule_load(ts_connection *c);

static pthread_mutex_t load_mutex = PTHREAD_MUTEX_INITIALIZER;

static unsigned long long uid = 0; /* A unique(ish) id to append to genid's to
                                    * help avoid clashses */

int ts_bulk_cache_flush(ts_connection *c, int verbosity, FILE *msg)
{
    int i, j, ret;
    char tmp[TMP_SIZE];

    for (i=TB; i<2; i++) {
	for (j=0; j<TS_SYM_SUBBLOCKS; j++) {
	    snprintf(tmp, TMP_SIZE-1, CACHE_FILE_TEMPLATE, c->db_name,
		     cache_letters[i], j);
	    ret = unlink(tmp);
	    if (ret && errno != ENOENT) {
		fprintf(msg, "Error removing cache file %s: %d - %s\n", 
			tmp, ret, sys_errlist[errno]);

		return 1;
	    }
	}
    }

    if (verbosity > 0) {
	fprintf(msg, "bulk import caches removed\n");
    }

    return 0;
}

int ts_bulk_import_init(ts_connection *c, int verbosity, FILE *msg)
{
    MYSQL_RES *res;
    MYSQL_ROW row;
    int i, j;
    int sync_cache = 0; /* true if we need to sync to the SQL cache */
    long long total, free;
    char tmp[TMP_SIZE];

    if (c->bid) {
	fprintf(msg, "error: bulk import allready initialised\n");

	return 1;
    }

    if (verbosity > 0) {
	fprintf(msg, "initialising bulk import\n");
    }

    ts_physical_ram(&total, &free);

    c->bid = calloc(1, sizeof(ts_bulk_import_data));

#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "SELECT MAX(tid) FROM "T_T);
    res = mysql_use_result(c->db);
    row = mysql_fetch_row(res);
    if (row && row[0]) {
	c->bid->tid = atoi(row[0])+1;
    } else {
	c->bid->tid = 1;
    }
    mysql_free_result(res);

    ts_mysql_query(c->db, "SELECT MAX(sid) FROM "T_S);
    res = mysql_use_result(c->db);
    row = mysql_fetch_row(res);
    if (row && row[0]) {
	c->bid->tid = atoi(row[0])+1;
    } else {
	c->bid->tid = 1;
    }
    mysql_free_result(res);
#else
    c->bid->dt_h = g_hash_table_new_full(g_str_hash, g_str_equal, ts_entry_free, NULL);
    c->bid->la_h = g_hash_table_new_full(g_str_hash, g_str_equal, ts_entry_free, NULL);

    ts_mysql_query(c->db, "SELECT id, uri FROM "T_D);
    res = mysql_use_result(c->db);
    while ((row = mysql_fetch_row(res))) {
	long id = atoi(row[0]);

	if (id > c->bid->max_dt) {
	    c->bid->max_dt = id;
	}
	g_hash_table_insert(c->bid->dt_h, strdup(row[1]), (gpointer)id);
    }
    mysql_free_result(res);

    ts_mysql_query(c->db, "SELECT id, lang FROM "T_L);
    res = mysql_use_result(c->db);
    while ((row = mysql_fetch_row(res))) {
	int id = atoi(row[0]);

	if (id > c->bid->max_la) {
	    c->bid->max_la = id;
	}
	g_hash_table_insert(c->bid->la_h, strdup(row[1]), (gpointer)atol(row[0]));
    }
    mysql_free_result(res);
#endif

    for (i=0; i<3; i++) {
	int ret;

#ifndef ENABLE_QUADS
	if (i != M) {
#else
	if (i != M && i != T) {
#endif
	    struct stat sbuf;

	    for (j=0; j<TS_SYM_SUBBLOCKS; j++) {
		db_create((c->bid->db[i])+j, NULL, 0);
		if (!c->bid->db[i]) {
		    fprintf(stderr, "fatal error: failed to create bdb handle %d\n",
			    i+1);

		    return -1;
		}

		/* create a reasonable cache, increasing this makes things go
		   quicker, until you run out of ram */
#ifndef ENABLE_QUADS
		c->bid->db[i][j]->set_cachesize(c->bid->db[i][j], 0, total/5, 1);
#else
		c->bid->db[i][j]->set_cachesize(c->bid->db[i][j], 0, total/3, 1);
#endif
		snprintf(tmp, TMP_SIZE-1, CACHE_FILE_TEMPLATE, c->db_name,
			 cache_letters[i], j);
		ret = stat(tmp, &sbuf);
		if (ret == -1 && errno == ENOENT) {
		    sync_cache = 1;
		}
		
		ret = c->bid->db[i][j]->open(c->bid->db[i][j], NULL, tmp, NULL,
					    DB_BTREE, DB_CREATE, 0666);
		if (ret) {
		    c->bid->db[i][j]->err(c->bid->db[i][j], ret, "bdb open error %d", ret);

		    return -2;
		}
	    }
	}

	snprintf(tmp, TMP_SIZE-1, LIVE_DAT_FILE_TEMPLATE, c->db_name, cache_letters[i]);
	c->bid->filename_live[i] = strdup(tmp);
	snprintf(tmp, TMP_SIZE-1, DAT_FILE_TEMPLATE, c->db_name, cache_letters[i]);
	c->bid->filename[i] = strdup(tmp);
    }
#ifdef ENABLE_QUADS
    for (j=1; j<TS_SYM_SUBBLOCKS; j++) {
	snprintf(tmp, TMP_SIZE-1, SUBBLOCK_FILE_TEMPLATE, c->db_name, cache_letters[S], j);
	c->bid->sb_filename[j] = strdup(tmp);
    }
#endif

    /* this lock is not optional, if its not taken then we might get
     * inconsitencies from other processes chaning the AI columns after
     * we read */
#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "LOCK TABLES "T_T" WRITE, "T_S" WRITE, "
		   T_I" WRITE, "T_F" WRITE, "T_A" WRITE, "T_M" WRITE");
#else
    ts_mysql_query(c->db, "LOCK TABLES "T_T" WRITE, "T_S" WRITE, "
		   T_I" WRITE, "T_F" WRITE, "T_A" WRITE, "T_M" WRITE"
		   T_D" WRITE, "T_L" WRITE");
#endif

    if (sync_cache) {
	DBT dk, dv;
#ifndef ENABLE_QUADS
	int32_t t, slot[3];
#endif
	ts_hash hash;
	int sb;
	int ret;
	int count = 0;

	TIME(NULL);
	if (verbosity > 0) {
	    fprintf(msg, "syncing database with integrity cache\n");
	}
	memset(&dk, 0, sizeof(dk));
	memset(&dv, 0, sizeof(dv));

#ifndef ENABLE_QUADS
	dk.data = slot;
	dk.size = sizeof(slot);
	dv.data = &t;
	dv.size = sizeof(t);
	ts_mysql_query(c->db, "SELECT tid, subject, predicate, object FROM "T_T);
	res = mysql_use_result(c->db);
	count = 0;
	while ((row = mysql_fetch_row(res))) {
	    t = atoi(row[0]);
	    slot[0] = atoi(row[1]);
	    slot[1] = atoi(row[2]);
	    slot[2] = atoi(row[3]);

	    ret = c->bid->db[T][0]->put(c->bid->db[T][0], NULL, &dk, &dv, 0);
	    if (ret) {
		c->bid->db[T][0]->err(c->bid->db[T][0], ret, "bdb put error %d", ret);
	    }
	    if (++count % 1000000 == 0 && verbosity > 0) {
		fprintf(msg, "synced %d triples\r", count);
		fflush(msg);
	    }
	}
	mysql_free_result(res);
	if (verbosity > 0) {
	    fprintf(msg, "synced %d triples\n", count);
	}
	TIME("cache_sync_triples");

	dk.data = &hash;
	dk.size = sizeof(hash);
	dv.data = &t;
	dv.size = sizeof(t);
	ts_mysql_query(c->db, "SELECT sid, hash FROM "T_S);
	res = mysql_use_result(c->db);
	count = 0;
	while ((row = mysql_fetch_row(res))) {
	    t = atoi(row[0]);
	    hash = atoll(row[1]);

	    ret = c->bid->db[S][0]->put(c->bid->db[S][0], NULL, &dk, &dv, 0);
	    if (ret) {
		c->bid->db[S][0]->err(c->bid->db[S][0], ret, "bdb put error %d", ret);
	    }
	    if (++count % 100000 == 0 && verbosity > 0) {
		fprintf(msg, "synced %d symbols\r", count);
		fflush(msg);
	    }
	}
	mysql_free_result(res);
	if (verbosity > 0) {
	    fprintf(msg, "synced %d symbols\n", count);
	}
	TIME("cache_sync_symbols");
#else
	dk.data = &hash;
	dk.size = sizeof(hash);
	dv.data = 0;
	dv.size = 0;
	ts_mysql_query(c->db, "SELECT hash FROM "T_S);
	res = mysql_use_result(c->db);
	count = 0;
	while ((row = mysql_fetch_row(res))) {
	    hash = atoll(row[0]);
	    sb = hash & (TS_SYM_SUBBLOCKS - 1);

	    ret = c->bid->db[S][sb]->put(c->bid->db[S][sb], NULL, &dk, &dv, 0);
	    if (ret) {
		c->bid->db[S][sb]->err(c->bid->db[S][sb], ret, "bdb put error %d", ret);
	    }
	    if (++count % 100000 == 0 && verbosity > 0) {
		fprintf(msg, "synced %d symbols\r", count);
		fflush(msg);
	    }
	}
	mysql_free_result(res);
	if (verbosity > 0) {
	    fprintf(msg, "synced %d symbols\n", count);
	}
	TIME("cache_sync_symbols");
#endif
    }

    TIME(NULL);
    ts_mysql_query(c->db, "ALTER TABLE "T_T" DISABLE KEYS");
    ts_mysql_query(c->db, "ALTER TABLE "T_S" DISABLE KEYS");
#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "ALTER TABLE "T_M" DISABLE KEYS");
#endif
    TIME("disable_keys");

    return 0;
}

int ts_bulk_import(ts_connection *c, char *uri, char *model, int verbosity,
    FILE *msg, ts_syntax format, int last)
{
    char timestamp[TMP_SIZE];
    time_t t;
    struct tm *tp;
    char *poll_obj_uri;
    int i;
    //ts_hash sys_model;
    raptor_parser *rdf_parser = NULL;
    raptor_uri ruri, muri;
    ts_bulk_parse_stuff parse_data;
    DB *dbm_t;
    GSList *it;

    dbm_t = c->bid->db[T][0];

    parse_data.verbosity = verbosity;
    parse_data.db = c;
    parse_data.msg = msg;
    parse_data.tax_rebuild = 0;
    parse_data.count_err = 0;
    parse_data.count_warn = 0;
    parse_data.count_trip = 0;
    memset(parse_data.checked_classes, 0, CLASS_H_SIZE * sizeof(ts_hash));

    if (!uid) {
	struct timeval tv;

	gettimeofday(&tv, NULL);
	uid = (unsigned int) getpid();
	uid ^= ((unsigned long long) tv.tv_usec) << 16;
	uid ^= ((unsigned long long) tv.tv_sec) << 32;
    } else {
	uid++;
    }
    poll_obj_uri = g_strdup_printf("3store:po:%llx", uid);

    t = time(NULL);
    tp = localtime(&t);
    snprintf(timestamp, 255, "%04d-%02d-%02dT%02d:%02d:%02d", tp->tm_year +
	     1900, tp->tm_mon + 1, tp->tm_mday, tp->tm_hour, tp->tm_min,
	     tp->tm_sec);

    for (i=0; i<3; i++) {
	c->bid->filep[i] = fopen(c->bid->filename_live[i], "w");
	if (!c->bid->filep[i]) {
	    fprintf(stderr, "Failed to create raw output file %s\n",
		     c->bid->filename_live[i]);

	    return 2;
	}
    }

#ifdef ENABLE_QUADS
    for (i=1; i<TS_SYM_SUBBLOCKS; i++) {
	/* If we've lcosed the file because we've emptied it, reopen it */
	if (!c->bid->sb_filep[i]) {
	    c->bid->sb_filep[i] = fopen(c->bid->sb_filename[i], "w");
	}
	if (!c->bid->sb_filep[i]) {
	    fprintf(stderr, "Failed to create raw buffer file %s\n",
		     c->bid->sb_filename[i]);

	    return 2;
	}
    }

    c->bid->subblock = 0;
#endif

    if (format == TS_AUTO) {
	if (strstr(uri, ".n3") || strstr(uri, ".ttl")) {
	    if (verbosity > 1) fprintf(msg, "using Turtle parser\n");
	    rdf_parser = raptor_new_parser("turtle");
	} else if (strstr(uri, ".nt")) {
	    if (verbosity > 1) fprintf(msg, "using NTriples parser\n");
	    rdf_parser = raptor_new_parser("ntriples");
	} else {
	    if (verbosity > 1) fprintf(msg, "using RDF/XML parser\n");
	    rdf_parser = raptor_new_parser("rdfxml");
	}
    } else if (format == TS_XML) {
	if (verbosity > 1) fprintf(msg, "using RDF/XML parser\n");
	rdf_parser = raptor_new_parser("rdfxml");
    } else if (format == TS_NTRIPLES) {
	if (verbosity > 1) fprintf(msg, "using NTriples parser\n");
	rdf_parser = raptor_new_parser("ntriples");
    } else if (format == TS_TURTLE) {
	if (verbosity > 1) fprintf(msg, "using Turtle parser\n");
	rdf_parser = raptor_new_parser("turtle");
    }
    if (!rdf_parser) {
	fprintf(msg, "Failed to create parser\n");
	return 1;
    }

    raptor_set_statement_handler(rdf_parser, &parse_data, store_stmt);
    raptor_set_warning_handler(rdf_parser, &parse_data, warning_handler);
    raptor_set_error_handler(rdf_parser, &parse_data, err_handler);
    raptor_set_fatal_error_handler(rdf_parser, &parse_data, fatal_err_handler);

    ruri = raptor_new_uri((unsigned char *)uri);

    if (!strcmp(model, "background")) {
	if (!c->feature[TS_FEATURE_STD_COMP]) {
	    fprintf(stderr, "Standards compliance mode is off, background graphs not supported\n");

	    return 1;
	}
	muri = raptor_new_uri((unsigned char *)"background:");
	parse_data.model = 0;
    } else {
	muri = raptor_new_uri((unsigned char *)model);
	parse_data.model = ts_hash_uri(model);
    }

    c->bid->model = check_symbol(c, parse_data.model, model, 1, 0);

    /* Process RDF input */
    if (raptor_parse_uri(rdf_parser, ruri, muri)) {
	fprintf(msg, "warning: failed to parse file correctly\n");
    }
    raptor_free_parser(rdf_parser);

    free(ruri);
    free(muri);

    TIME(NULL);
    wait_load();
    TIME("db_write_thread_wait");

    /* insert pending statements that had to wait for DB lock */
    for (it = c->bid->pending_sql; it; it = it->next) {
	ts_mysql_query(c->db, it->data);
	free(it->data);
	it->data = NULL;
    }
    g_slist_free(c->bid->pending_sql);
    c->bid->pending_sql = NULL;

    if (parse_data.model && c->feature[TS_FEATURE_PROVENANCE]) {
	ts_id file_model;
	ts_id sys_model = TS_ID(c, ts_hash_uri(TS_SYS_MODEL));
	char *query;

	if (verbosity > 1) {
	    printf("Asserting provenance\n");
	}
	sys_model = check_symbol(c, ts_hash_uri(TS_SYS_MODEL), TS_SYS_MODEL,
				 c->uri_dt, 0);
#ifndef ENABLE_QUADS
	query = g_strdup_printf("DELETE FROM "T_T" WHERE "
				"tid=(SELECT tid FROM "T_M" WHERE mid=%lld) "
				"AND subject=%lld AND predicate=%lld",
			TS_ID(c, sys_model), TS_ID(c, c->model),
			TS_ID(c, ts_hash_uri(INT_NS "last-polled")));
#else /* ENABLE_QUADS */
	query = g_strdup_printf("DELETE FROM "T_T" WHERE "
				"model=%lld AND subject=%lld AND "
				"predicate=%lld",
			sys_model, c->model,
			ts_hash_uri(INT_NS "last-polled"));
#endif /* ENABLE_QUADS */
	if (mysql_query(c->db, query)) {
	    fprintf(msg, "Warning: failed to remove old timestamp '%s' (%d): "
		    "%s\n", query, mysql_errno(c->db), mysql_error(c->db));
	}
	free(query);

	file_model = c->bid->model;
	c->bid->model = sys_model;
	ts_bulk_assert_triple(c, &parse_data, model,
		         "http://purl.org/dc/elements/1.1/source", uri, ObjURI,
			 0, 0);
	ts_bulk_assert_triple(c, &parse_data, model, INT_NS "last-polled",
		         poll_obj_uri, ObjURI, 0, 0);
	ts_bulk_assert_triple(c, &parse_data, model, INT_NS "polled", poll_obj_uri,
		         ObjURI, 0, 0);
	ts_bulk_assert_triple(c, &parse_data, poll_obj_uri, INT_NS "date", timestamp,
			 ObjLiteral, 0, 0);
	if (parse_data.count_err == 0 && parse_data.count_warn == 0) {
	    ts_bulk_assert_triple(c, &parse_data, poll_obj_uri,
		             INT_NS "status-message", "OK", ObjLiteral, 0, 0);
	    ts_bulk_assert_triple(c, &parse_data, poll_obj_uri, INT_NS "status",
			     INT_NS "success", ObjURI, 0, 0);
	} else {
	    char msg[TMP_SIZE];

	    snprintf(msg, 255, "%d error%s and %d warning%s",
		     parse_data.count_err, parse_data.count_err == 1 ? "" : "s",
		     parse_data.count_warn, parse_data.count_warn == 1 ?
		                            "" : "s");
	    ts_bulk_assert_triple(c, &parse_data, poll_obj_uri,
		             INT_NS "status-message", msg, ObjLiteral, 0, 0);
	    if (parse_data.count_err > 0) {
		ts_bulk_assert_triple(c, &parse_data, poll_obj_uri, INT_NS "status",
				 INT_NS "failure", ObjURI, 0, 0);
	    } else {
		ts_bulk_assert_triple(c, &parse_data, poll_obj_uri, INT_NS "status",
				 INT_NS "warning", ObjURI, 0, 0);
	    }
	}
	c->bid->model = file_model;
	TIME("update_provenance");
    }
    free(poll_obj_uri);

#ifdef ENABLE_QUADS
    if (c->bid->sb_count > 1000000 || last) {
	if (verbosity > 1) {
	    printf("Asserting symbol sub-blocks\n");
	}
	TIME(NULL);
	c->bid->sb_count = 0;
	/* run through delayed possible symbols imports */
	for (i=1; i<TS_SYM_SUBBLOCKS; i++) {
	    c->bid->subblock = i;
	    fclose(c->bid->sb_filep[i]);
	    c->bid->sb_filep[i] = fopen(c->bid->sb_filename[i], "r");
	    if (!c->bid->sb_filep[i]) {
		fprintf(stderr, "Failed to reopen raw buffer file %s\n",
			 c->bid->sb_filename[i]);

		return 2;
	    }
	    parse_subblock(c, i);
	    fclose(c->bid->sb_filep[i]);
	    c->bid->sb_filep[i] = NULL;
	}
	TIME("add_subblocks");
    }
#endif

    /* close and flip files for loading */
    for (i=0; i<3; i++) {
	fclose(c->bid->filep[i]);
	if (rename(c->bid->filename_live[i], c->bid->filename[i])) {
	    fprintf(msg, "Error moving data file %s: %s\n", 
	            c->bid->filename[i], sys_errlist[errno]);

	    return 1;
	}
    }
    schedule_load(c);

    if (parse_data.count_err) {
	return -parse_data.count_err;
    } else {
	return parse_data.count_trip;
    }
}

void *load_thread(ts_connection *c)
{
    char tmp[TMP_SIZE];
#ifdef TIMING
    struct timeval then, now;
    double diff;
#endif

    pthread_mutex_lock(&load_mutex);

#ifdef TIMING
    gettimeofday(&then, 0);
#endif
    snprintf(tmp, TMP_SIZE-1, "LOAD DATA LOCAL INFILE '%s' IGNORE INTO TABLE "T_T,
	     c->bid->filename[T]);
    ts_mysql_query(c->db, tmp);
#ifdef TIMING
    gettimeofday(&now, 0);
    diff = (now.tv_sec - then.tv_sec) + (now.tv_usec - then.tv_usec) * 0.000001;
    ts_timing_add("bulk_insert_triples", diff, -1);
#endif

#ifdef TIMING
    gettimeofday(&then, 0);
#endif
    snprintf(tmp, TMP_SIZE-1, "LOAD DATA LOCAL INFILE '%s' IGNORE INTO TABLE "T_S,
	     c->bid->filename[S]);
    ts_mysql_query(c->db, tmp);
#ifdef TIMING
    gettimeofday(&now, 0);
    diff = (now.tv_sec - then.tv_sec) + (now.tv_usec - then.tv_usec) * 0.000001;
    ts_timing_add("bulk_insert_symbols", diff, -1);
#endif

#ifndef ENABLE_QUADS
#ifdef TIMING
    gettimeofday(&then, 0);
#endif
    snprintf(tmp, TMP_SIZE-1, "LOAD DATA LOCAL INFILE '%s' IGNORE INTO TABLE "T_M,
	     c->bid->filename[M]);
    ts_mysql_query(c->db, tmp);
#ifdef TIMING
    gettimeofday(&now, 0);
    diff = (now.tv_sec - then.tv_sec) + (now.tv_usec - then.tv_usec) * 0.000001;
    ts_timing_add("bulk_insert_tmem", diff, -1);
#endif
#endif

    pthread_mutex_unlock(&load_mutex);

    return NULL;
}

static void parse_subblock(ts_connection *c, int block)
{
    FILE *f = c->bid->sb_filep[c->bid->subblock];
    char *ret;
    int reti;
    char buffer[32768];
    ts_hash hash;
    DB *d = c->bid->db[S][c->bid->subblock];
    DBT dk, dv;

    memset(&dk, 0, sizeof(dk));
    memset(&dv, 0, sizeof(dv));

    dk.size = sizeof(hash);
    dk.data = &hash;

    while ((ret = fgets(buffer, sizeof(buffer)-1, f))) {
	if (ret == NULL) {
	    break;
	}
	reti = sscanf(buffer, "%lld", &hash);

	if (d->get(d, NULL, &dk, &dv, 0) == DB_NOTFOUND) {
	    reti = d->put(d, NULL, &dk, &dv, 0);
	    fputs(buffer, c->bid->filep[S]);
	}
    }
}

static pthread_t loader = 0;

/* block until the loading worker thread has finished */
static void wait_load()
{
    static int first = 1;

    if (!first) {
	pthread_join(loader, NULL);
    }
    first = 0;
}

static void schedule_load(ts_connection *c)
{
    int ret;

    ret = pthread_create(&loader, NULL, (void*)&load_thread, c);
    if (ret) {
	printf("thread error %d\n", ret);
    }
}

int ts_bulk_import_commit(ts_connection *c, int verbosity, FILE *msg)
{
    int i, j;

    if (!c->bid) {
	fprintf(msg, "bulk import not initialised\n");

	return -1;
    }

    /* this makes sure that the last loader thread has finished */
    if (loader) pthread_join(loader, NULL);

    if (verbosity > 0) {
	fprintf(msg, "commiting bulk import\n");
    }
    TIME(NULL);

    for (i=TB; i<2; i++) {
	for (j=0; j<TS_SYM_SUBBLOCKS; j++) {
	    if (c->bid->db[i][j]) {
		c->bid->db[i][j]->close(c->bid->db[i][j], 0);
	    }
	}
    }

    ts_mysql_query(c->db, "ALTER TABLE "T_T" ENABLE KEYS");
    ts_mysql_query(c->db, "ALTER TABLE "T_S" ENABLE KEYS");
#ifndef ENABLE_QUADS
    ts_mysql_query(c->db, "ALTER TABLE "T_M" ENABLE KEYS");
#endif
    TIME("enable_keys");
    
    ts_mysql_query(c->db, "UNLOCK TABLES");

    for (i=0; i<3; i++) {
	free(c->bid->filename[i]);
	free(c->bid->filename_live[i]);
    }
    if (c->bid->dt_h) {
	g_hash_table_destroy(c->bid->dt_h);
    }
    if (c->bid->la_h) {
	g_hash_table_destroy(c->bid->la_h);
    }
    free(c->bid);
    c->bid = NULL;

    return 0;
}

static void store_stmt(void *user_data, const raptor_statement *statement)
{
    ts_bulk_parse_stuff *data = (ts_bulk_parse_stuff *)user_data;
    ts_obj_type type;
    char tmps[TMP_SIZE], tmpp[TMP_SIZE], tmpo[TMP_SIZE];
    char *subj = (char *)raptor_uri_as_string((raptor_uri *)statement->subject),
	 *pred = (char *)raptor_uri_as_string((raptor_uri *)statement->predicate),
         *obj;
    const char *datatype = (char *)
	raptor_uri_as_string(statement->object_literal_datatype);
    long dt = 0;
    long lang = 0;

    if (datatype) {
#ifndef ENABLE_QUADS
	dt = check_symbol(data->db, ts_hash_uri(datatype), (char *)datatype, 
			  data->db->uri_dt, 0);
#else
	dt = (long)g_hash_table_lookup(data->db->bid->dt_h, (char *)datatype);
	if (!dt) {
	    char *q;

	    data->db->bid->max_dt++;
	    dt = data->db->bid->max_dt;
	    //XXX needs escaping
	    q = g_strdup_printf("INSERT INTO "T_D" VALUES(%ld, '%s')", dt, (char *)datatype);
	    data->db->bid->pending_sql = g_slist_prepend(data->db->bid->pending_sql, q);
	    g_hash_table_insert(data->db->bid->dt_h, strdup(datatype), (gpointer)dt);
	}
#endif
    }
    if (statement->object_literal_language) {
	char *langtag = (char *)statement->object_literal_language;

#ifndef ENABLE_QUADS
	lang = check_symbol(data->db, ts_hash_literal(langtag, 0, 0), langtag, 0, 0);
#else
	lang = (long)g_hash_table_lookup(data->db->bid->dt_h, langtag);
	if (!lang) {
	    char *q;

	    data->db->bid->max_la++;
	    lang = data->db->bid->max_la;
	    //XXX needs escaping
	    q = g_strdup_printf("INSERT INTO "T_L" VALUES(%ld, '%s')", lang, langtag);
	    data->db->bid->pending_sql = g_slist_prepend(data->db->bid->pending_sql, q);
	    g_hash_table_insert(data->db->bid->la_h, langtag, (gpointer)lang);
	}
#endif
    }

    /* pick a bnode suffix, if we have a model ID, use that, otherwise (ie.
     * were in the background graph, usea unique-ish int */
    long long bnodesuf = data->model ? data->model : uid;

    (data->count_trip)++;
    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_LITERAL ||
	statement->object_type == RAPTOR_IDENTIFIER_TYPE_XML_LITERAL) {
	type = ObjLiteral;
	obj = (char *)statement->object;
    } else {
	type = ObjURI;
	obj = (char *)raptor_uri_as_string((raptor_uri *)statement->object);
	dt = data->db->uri_dt;
    }

    if (statement->subject_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmps, "_:%s.%llx", (char *)statement->subject, bnodesuf);
	subj = tmps;
    }
    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmpp, "_:%s.%llx", (char *)statement->predicate, bnodesuf);
	pred = tmpp;
    }
    if (statement->object_type == RAPTOR_IDENTIFIER_TYPE_ANONYMOUS) {
	sprintf(tmpo, "_:%s.%llx", (char *)statement->object, bnodesuf);
	obj = tmpo;
    }

    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
	sprintf(tmpp, MEMBER_PREFIX "%d", *((int *)pred));
	pred = tmpp;
    }

    if (data->verbosity > 2) {
	fprintf(stdout, "asserting (<%s>, <%s>, %c%s%c), dt:%s, lang:%s\n",
                subj, pred,
	        type == ObjLiteral ? '"' : '<', obj,
	        type == ObjLiteral ? '"' : '>', (char *)datatype,
		(char *)statement->object_literal_language);
    }

    ts_bulk_assert_triple(data->db, data, subj, pred, obj, type, dt, lang);

    if (statement->predicate_type == RAPTOR_IDENTIFIER_TYPE_ORDINAL) {
	ts_bulk_assert_triple(data->db, data, subj, MEMBER_PROP, obj,
	    type, dt, lang);
	ts_bulk_assert_triple(data->db, data, pred, RDFS_SUBPROPERTY,
			     MEMBER_PROP, ObjURI, 0, 0);
    }

#if 0
    /* checks for pending inference */
    if (data->db->feature[TS_FEATURE_INFERENCE] && !data->tax_rebuild) {
	if (!strcmp(pred, RDFS_SUBCLASS) || !strcmp(pred, RDFS_SUBPROPERTY)) {
	    data->tax_rebuild = 1;
	} else if (!strcmp(pred, RDF_TYPE)) {
	    ts_hash class = ts_hash_uri(obj);

	    /* if we may not have seen this class */
	    if (data->checked_classes[class & (CLASS_H_SIZE - 1)] != class) {
		char tmp[TMP_SIZE];
		int count;

		data->checked_classes[class & (CLASS_H_SIZE - 1)] = class;
		snprintf(tmp, TMP_SIZE, "SELECT 1 FROM "T_T" WHERE "
			 "subject=%lld AND predicate=-5275523445234180533 AND "
			 "object=%lld LIMIT 1", class, class);
		count = ts_sql_count(data->db->db, tmp);
		if (count == 0) {
		    ts_hash tc_hash = TS_ID(data->db, ts_ensure_resource
			    (data->db, INT_CLOSURE));
		    /* shortcut the taxonomic closure code incase theres no
		     * class graph. If the taxonomic closure is recalculated
		     * later this will be overwritten anyway */
		    ts_assert_triple(data->db, tc_hash, obj, RDFS_SUBCLASS,
				     obj, ObjURI, 0, 2);
		    ts_assert_triple(data->db, tc_hash, obj, RDF_TYPE,
				     RDFS_CLASS, ObjURI, 0, 2);
		}
	    }
	}
    }
#endif
}

#ifndef ENABLE_QUADS
static void ts_bulk_assert_triple(ts_connection *c, ts_bulk_parse_stuff *data,
				  char *subj, char *pred, char *obj, 
				  ts_obj_type type, int dt, int lang)
{
    DBT dkt, dvt;
    DB *dbm_t;
    int32_t kdata[3];
    int ret;

    TIME(NULL);
    memset(&dkt, 0, sizeof(dkt));
    memset(&dvt, 0, sizeof(dvt));

    dbm_t = c->bid->db[T];

    kdata[0] = check_symbol(c, ts_hash_uri(subj), subj, 1, 0);
    kdata[1] = check_symbol(c, ts_hash_uri(pred), pred, 1, 0);
    if (type == ObjURI) {
        kdata[2] = check_symbol(c, ts_hash_uri(obj), obj, 1, 0);
    } else {
        kdata[2] = check_symbol(c, ts_hash_literal(obj, dt, lang), obj,
			        dt, lang);
    }
    dkt.data = kdata;
    dkt.size = sizeof(kdata);

    ret = dbm_t->get(dbm_t, NULL, &dkt, &dvt, 0);
    if (ret == DB_NOTFOUND) {
	dvt.data = &(c->bid->tid);
	dvt.size = sizeof(int32_t);
	ret = dbm_t->put(dbm_t, NULL, &dkt, &dvt, 0);
	if (ret) {
	    dbm_t->err(dbm_t, ret, "bdb put error %d", ret);
	}
	fprintf(c->bid->filep[T], "%d\t%d\t%d\t%d\n", c->bid->tid, kdata[0],
		kdata[1], kdata[2]);
    } else if (ret != 0) {
	dbm_t->err(dbm_t, ret, "bdb get error %d", ret);
    }

    fprintf(c->bid->filep[M], "%d\t%d\n", c->bid->tid, c->bid->model);
    c->bid->tid++;
    TIME("bulk_bdb_integrity_check");
}
#else
static void ts_bulk_assert_triple(ts_connection *c, ts_bulk_parse_stuff *data,
				  char *subj, char *pred, char *obj, 
				  ts_obj_type type, int dt, int lang)
{
    ts_id s, p, o;

    TIME(NULL);
    s = check_symbol(c, ts_hash_uri(subj), subj, c->uri_dt, 0);
    p = check_symbol(c, ts_hash_uri(pred), pred, c->uri_dt, 0);
    if (type == ObjURI) {
        o = check_symbol(c, ts_hash_uri(obj), obj, c->uri_dt, 0);
    } else {
        o = check_symbol(c, ts_hash_literal(obj, dt, lang), obj, dt, lang);
    }

    fprintf(c->bid->filep[T], "%lld\t%lld\t%lld\t%lld\n", c->bid->model, s,
		p, o);
    TIME("bulk_triple_log_append");
}
#endif

#define HASH_SIZE 4096

static int symbol_cache_l1_hit = 0;
static int symbol_cache_l2_hit = 0;
static int symbol_cache_miss = 0;

void ts_bulk_import_get_cache_stats(int *l1, int *l2, int *miss)
{
    *l1 = symbol_cache_l1_hit;
    *l2 = symbol_cache_l2_hit;
    *miss = symbol_cache_miss;
}

static ts_hash cache_h[HASH_SIZE];
#ifndef ENABLE_QUADS
static ts_id   cache_i[HASH_SIZE];
#endif

static ts_id check_symbol(ts_connection *c, ts_hash hash, char *lex,
			    int dt, int lang)
{
    DB *d;
    DBT dk, dv;
    int32_t id;
    int ret;
#ifdef ENABLE_QUADS
    char *ebv;
    int sb;
#endif

#ifndef ENABLE_QUADS
    d = c->bid->db[S][0];
#else
    d = c->bid->db[S][hash & (TS_SYM_SUBBLOCKS - 1)];
#endif

    if (cache_h[hash & (HASH_SIZE-1)] == hash) {
	symbol_cache_l1_hit++;

#ifndef ENABLE_QUADS
	return cache_i[hash & (HASH_SIZE-1)];
#else
	return hash;
#endif
    }

#ifdef ENABLE_QUADS
    sb = hash & (TS_SYM_SUBBLOCKS - 1);
    if (sb != c->bid->subblock) {
	return check_symbol_later(c, hash, lex, dt, lang);
    }
#endif

    memset(&dk, 0, sizeof(dk));
    memset(&dv, 0, sizeof(dv));

    dk.size = sizeof(hash);
    dk.data = &hash;
    if (d->get(d, NULL, &dk, &dv, 0) == DB_NOTFOUND) {
	symbol_cache_miss++;
#ifndef ENABLE_QUADS
        id = c->bid->sid++;
        dv.size = sizeof(id);
        dv.data = &id;
#else
	id = 1;
        dv.size = 0;
        dv.data = 0;
#endif
        ret = d->put(d, NULL, &dk, &dv, 0);
        if (ret) {
            d->err(d, ret, "put error %d", ret);
        }
#ifndef ENABLE_QUADS
	fprintf(c->bid->filep[S], "%u\t%lld\t%s\t%d\t%d\n", c->bid->sid,
		hash, lex, dt, lang);
#else
	ebv = ts_insert_symbol_values(c, &(c->bid->pending_sql), hash, lex, dt,
				      lang);
	fprintf(c->bid->filep[S], "%lld\t%s\t%s\t%d\t%d\n", hash, lex, ebv,
		dt, lang);
#endif
	cache_h[hash & (HASH_SIZE-1)] = hash;
#ifndef ENABLE_QUADS
	cache_i[hash & (HASH_SIZE-1)] = id;

        return id;
#else
        return hash;
#endif
    } else {
	symbol_cache_l2_hit++;
	cache_h[hash & (HASH_SIZE-1)] = hash;
#ifndef ENABLE_QUADS
	cache_i[hash & (HASH_SIZE-1)] = *(int32_t *)(dv.data);

        return *(int32_t *)(dv.data);
#else

        return hash;
#endif
    }
}

static ts_id check_symbol_later(ts_connection *c, ts_hash hash, char *lex,
			    int dt, int lang)
{
    int sb = hash & (TS_SYM_SUBBLOCKS - 1);

    c->bid->sb_count++;
    fprintf(c->bid->sb_filep[sb], "%lld\t%s\t\%d\t%d\n", hash, lex, dt, lang);
    cache_h[hash & (HASH_SIZE-1)] = hash;

    return hash;
}

static void
fatal_err_handler(void *user_data, raptor_locator * locator,
                    const char *message)
{
    ts_bulk_parse_stuff *data = (ts_bulk_parse_stuff *)user_data;

    fprintf(data->msg, "3store import fatal error: ");
    raptor_print_locator(data->msg, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
err_handler(void *user_data, raptor_locator * locator, const char *message)
{
    ts_bulk_parse_stuff *data = (ts_bulk_parse_stuff *)user_data;

    fprintf(data->msg, "3store import error: ");
    raptor_print_locator(data->msg, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_err)++;
}

static void
warning_handler(void *user_data, raptor_locator * locator, const char *message)
{
    ts_bulk_parse_stuff *data = (ts_bulk_parse_stuff *)user_data;

    fprintf(data->msg, "3store import warning: ");
    raptor_print_locator(stderr, locator);
    fprintf(data->msg, " - %s\n", message);
    (data->count_warn)++;
}

/* vi:set ts=8 sts=4 sw=4: */
