#ifndef DATATYPES_H
#define DATATYPES_H

#include <glib.h>
#include <mysql.h>
#include <stdlib.h>
#include <stdint.h>
#include <limits.h>
#include <db.h>

#include "config.h"

/* the maximum number of blocks in a query */
#define TS_MAX_BLOCKS  256
#define TS_ORDER_BLOCK (TS_MAX_BLOCKS-1)
#define TS_MAX_TRIPLES 256

#ifdef ENABLE_QUADS
#define TS_SYM_SUBBLOCKS 4
#else
#define TS_SYM_SUBBLOCKS 1
#endif

/* supported RDF syntaxes */ 
typedef enum {
        TS_AUTO = 0,
        TS_XML,
        TS_NTRIPLES,
        TS_TURTLE 
} ts_syntax;

/* suported pre-kb features */
typedef enum {
        TS_FEATURE_STD_COMP = 0,
        TS_FEATURE_INFERENCE,
        TS_FEATURE_PROVENANCE,
        TS_FEATURE_LOADING,
	TS_FEATURE_LAST = TS_FEATURE_LOADING
} ts_feature;

extern char *ts_feature_names[TS_FEATURE_LAST+1];

/* types that objects can take */
typedef enum {
        ObjAny = -1,
        ObjURI = 0,
        ObjLiteral = 1
} ts_obj_type;

#if SIZEOF_LONG_LONG_INT == 8
typedef long long int ts_hash;
#else
typedef int64_t ts_hash;
#warning Cannot detect length of long long, you will gets lots of warnings
#endif

#ifndef ENABLE_QUADS
#define IDFS "%d"
typedef uint32_t ts_id;
#else
#define IDFS "%lld"
typedef ts_hash ts_id;
#endif

typedef struct ts_bulk_import_data_s {
	int32_t tid;		/* running triple id */
	int32_t sid;		/* running symbol id */
	ts_id model;		/* model id of current model */
	DB *db[3][TS_SYM_SUBBLOCKS];
				/* BDB handles for triples, sybols and tmem */
	char *filename[3];	/* mysql import files */
	char *filename_live[3];	/* mysql import files */
	char *sb_filename[TS_SYM_SUBBLOCKS];
				/* potential mysql import files for subblocks */
	FILE *filep[3];		/* file ptrs to live import files */
	FILE *sb_filep[TS_SYM_SUBBLOCKS];
				/* files to hold potentially missing symbols */
	int subblock;		/* number of current subblock that were
				 * processing */
	int sb_count;		/* number of symbols added since we synced */
	GHashTable *dt_h;	/* hash to hold datatype ids in quad mode */
	GHashTable *la_h;       /* hash to hold language ids in quad mode */
	int max_dt;		/* maximum datatype ID used */
	int max_la;		/* maximum language ID used */
	GSList *pending_sql;	/* SQL statements to execute after block */
} ts_bulk_import_data;

typedef struct ts_connection_s {
	void   *db;		/* the mysql DB object */
	char *db_name;
	ts_hash model;		/* hash of current model */
	int32_t uri_dt;		/* datatypes for well-known types */
	int32_t string_dt;
	int32_t int_dt;
	int32_t decimal_dt;
	int32_t float_dt;
	int32_t double_dt;
	int32_t datetime_dt;
	int32_t date_dt;
	int32_t time_dt;
	int32_t bool_dt;
	int     opt_lvl;	/* optimisation level */
	int feature[TS_FEATURE_LAST+1];
				/* what features are enabled for this store */
	int block_mode;		/* true if were using the pending table for
				 * assertions */
	int lock_tables;
	ts_bulk_import_data *bid;
} ts_connection;

typedef struct ts_binding_desc_s {
	char *sym;
	int col;
	int dt_col;
	int lang_col;
} ts_binding_desc;

typedef struct ts_binding_s {
	const char *sym;
	const char *lexical;
	const char *datatype;
	const char *lang;
} ts_binding;

typedef struct ts_query_s {
	rasqal_query *rquery;	/* the query that was used to build this one */
	raptor_uri *base;	/* base uri of query */
	GSList *glimit;		/* graphs that the reuslts must be drawn from */
	GSList *sql_setup;      /* sql commands to issue before querying */
	GSList *sql_pre_query;  /* sql commands to populate tmp tables etc. */
	GSList *sql_cleanup;	/* sql commands to remove tmp tables etc. */
	char *sql_query;	/* main sql query for results */
	char *error_str;	/* error str, error if set */
	int cols;		/* number of sparql reuslt columns */
	raptor_sequence *vars;  /* vars returned by the query */
	ts_binding_desc *col_desc;
				/* array of structs describing result cols */
	ts_binding *bindings;	/* place to store bound results per row */
	GSList *globe;		/* expressions that have to be appplied
				   globally */
	char *tt_names[TS_MAX_BLOCKS];
				/* the names of the intermediate tables */
	int ordered;		/* true if the results are ordered */
	int distinct;		/* true if the results are DISTINCT */
	struct ts_symtab_s *symtab; /* pointer to symtab for this query */
} ts_query;

typedef struct ts_row_s {
	MYSQL_ROW row;
	ts_binding_desc *col_desc;
	ts_binding *bindings;
	int cols;
} ts_row;

typedef struct ts_result_s {
	MYSQL_RES *res;
	ts_connection *con;
	ts_query *query;
	ts_row row;
	char *error_str;
} ts_result;

typedef struct ts_rel_query_s {
	char *map;
	char *project;
	char *join;
	char *select;
} ts_rel_query;

void ts_query_free(ts_query *q);

char *ts_rel_query_to_string(ts_rel_query *q, char *prefix, char *postfix);

void ts_rel_query_free(ts_rel_query *q);

#endif
