/*
 *  Copyright (C) 2005 Steve Harris
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: ts-query.c,v 1.13 2005/12/21 14:29:09 theno23 Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <rasqal.h>

#include "tstore.h"
#include "results.h"
#include "cmdline-parser.h"

/* write results to stdout based on format string */
void write_results(ts_result *res, const char *format, int headers, const char *stylesheet);

/* read stdin and run every query followed by #EOQ on a line of its own */
void programatic_io(ts_connection *con, const char *query_lang, const char *result_format);

int main(int argc, char*argv[])
{
    ts_connection *con;
    int help = 0;
    char *optstring = TSP_OPTSTR "l:f:O:ms:P";
    int c, opt_index = 0;
    static struct option long_options[] = {
        TSP_LONGOPTS,
        { "language",     1, 0, 'l' },
        { "format",       1, 0, 'f' },
        { "optimise",     1, 0, 'O' },
        { "mime-headers", 0, 0, 'm' },
        { "stylesheet",   1, 0, 's' },
        { "programatic",  0, 0, 'P' },
        { 0, 0, 0, 0 }
    };
    ts_result *res = NULL;
    char *query_string;
    char *query_lang = "sparql";
    char *result_format = "sparql";
    char *stylesheet = NULL;
    int headers = 0;
    int optimisation = -1;
    int programatic = 0;

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
        if (!tsp_handle_option(c, optarg)) {
            switch (c) {
	    case 'l':
		query_lang = optarg;
		break;
	    case 'f':
		result_format = optarg;
		break;
	    case 'O':
		optimisation = atoi(optarg);
		break;
	    case 'm':
		headers = 1;
		break;
	    case 's':
		stylesheet = optarg;
		break;
	    case 'P':
		programatic = 1;
		break;
            case 'h':
                help++;
                break;
            default:
                help++;
                break;
            }
        }
    }

    if (help || argc > optind + 1) {
	int i;
	const char *lang;
	const char *label;

	rasqal_init();

        fprintf(stderr, "Usage: %s "TSP_OPTIONS" [-l language] "
	        "[-f format] [-O level] [<query>]\n" TSP_HELP, basename(argv[0]));
	fprintf(stderr, "   -l, --language  use lanugage for this query\n");
	fprintf(stderr, "   -f, --result-format use format for this query\n");
	fprintf(stderr, "   -O, --optimsise set optimsiation level (default 1)\n");
	fprintf(stderr, "                   0=none, 1=normal, 2=high, 3=experimental\n");
	fprintf(stderr, "    where language is one of:");
	for (i=0; 1; i++) {
	    if (rasqal_languages_enumerate(i, &lang, &label, NULL)) {
		break;
	    }
	    fprintf(stderr, " %s", lang);
	}
	fprintf(stderr, "\n");
	fprintf(stderr, "    and format is one of:");
	for (i=0; 1; i++) {
	    if (!ts_serialisation_format_by_index(i)) {
		break;
	    }
	    fprintf(stderr, " %s", ts_serialisation_format_by_index(i));
	}
	fprintf(stderr, "\n\n");
	fprintf(stderr, "See man pages for more details\n");
	
        return 1;
    }

    ts_init();
    con = tsp_connect(NULL);

    if (programatic) {
	programatic_io(con, query_lang, result_format);
    } else if (argc > optind) {
	query_string = argv[optind];
	if (optimisation >= 0) ts_connection_set_optimisation(con, optimisation);
	ts_query *tq = ts_query_prepare(con, query_lang, query_string,
				      "3store:default#");
	res = ts_query_execute(con, tq);

	write_results(res, result_format, headers, stylesheet);

	ts_query_free(tq);
    }
    ts_disconnect(con);
    ts_finish();

    return 0;
}

#define MAX_Q_SIZE 16384

void programatic_io(ts_connection *con, const char *query_lang, const char *result_format)
{
    ts_result *res;
    ts_query *tq;
    char query[MAX_Q_SIZE];
    char *pos;
    char *newl;

    do {
	pos = query;
	*query = '\0';
	/* assemble query string */
	do {
	    newl = fgets(pos, query + MAX_Q_SIZE - pos - 1, stdin);
	    if (newl) {
		pos += strlen(newl);
	    }
	} while (newl && strcmp(newl, "#EOQ\n") && strcmp(newl, "#END\n"));
   
	/* process query string */ 
	if (*query && strcmp(query, "#EOQ\n") && strcmp(query, "#END\n")) {
	    tq = ts_query_prepare(con, query_lang, query, "file:stdin");
	    res = ts_query_execute(con, tq);
	    write_results(res, result_format, 1, NULL);
	    ts_query_free(tq);
	    if (!strcmp(result_format, "testcase") ||
	        !strcmp(result_format, "ascii")) {
		printf("# EOR\n");
	    } else {
		printf("<!-- EOR -->\n");
	    }
	    fflush(stdout);
	}
    } while (newl && strcmp(newl, "#END\n"));
}

void write_results(ts_result *res, const char *format, int headers, const char *stylesheet)
{
    if (!res) {
	return;
    }

    ts_serialise_results(res, format, (ts_outp)fprintf, stdout, headers,
			 stylesheet);
}

/* vi:set ts=8 sts=4 sw=4: */
