/*
 *  Copyright (C) 2002 Steve Harris, ECS, University of Southampton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: ts-setup.c,v 1.16 2005/11/16 18:46:51 theno23 Exp $
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <mysql.h>
#include <mysql_version.h>
#include <mysqld_error.h>
#include <errmsg.h>
#include <getopt.h>
#include <locale.h>
#include <unistd.h>
#include <termios.h>
#include <libgen.h>
#include <readline/readline.h>

#include "tstore.h"
#include "bulk-import.h"
#include "cmdline-parser.h"
#include "config.h"

void create_db(MYSQL *mdb);
void create_user(MYSQL *mdb);
void create_tables(MYSQL *mdb, int errase);
void erase_db(MYSQL *mdb);

static int verbosity = 0;
static int db_exists = 0;
static int dummy = 0;

#define LOG(l, a...) { if (verbosity >= l) printf(a); }

#define QUERY(con, qs) (dummy ? (printf("SQL: %s\n", qs), 0) : \
			        mysql_query(con, qs))

#define DEFAULT_FEATURES "IP"

int main(int argc, char *argv[])
{
    MYSQL *mdb;
    ts_connection fakecon;
    ts_connection *con;
    char *optstring = TSP_OPTSTR "vsef:n";
    int help = 0;
    int show_uris = 0;
    int erase = 0;
    int c, opt_index = 0;
    char *mysql_db;
    char tmp[TS_CF_VAL_SIZE];
    char *features = DEFAULT_FEATURES;
    int features_set = 0;
    int i;

    static struct option long_options[] = {
        TSP_LONGOPTS,
        { "verbose", 0, 0, 'v' },
        { "show", 0, 0, 's' },
        { "erase", 0, 0, 'e' },
        { "feature", 1, 0, 'f' },
        { "print-only", 0, 0, 'n' },
        { 0, 0, 0, 0 }
    };

    setlocale(LC_ALL, NULL);
    ts_init();

    while ((c = getopt_long (argc, argv, optstring, long_options, &opt_index)) != -1) {
        if (!tsp_handle_option(c, optarg)) {
            if (c == 'v') {
                verbosity++;
            } else if (c == 's') {
                show_uris = 1;
            } else if (c == 'e') {
                erase = 1;
            } else if (c == 'n') {
                dummy = 1;
	    } else if (c == 'f') {
		features = optarg;
		features_set = 1;
            } else {
                help++;
            }
        }
    }

    if (help) {
        fprintf(stderr, "Usage: %s "TSP_OPTIONS" [-v] [-s] [-e] [-f features] \n", basename(argv[0]));
        fprintf(stderr, TSP_HELP);
        fprintf(stderr, "   -v, --verbose     increase verbosity\n");
        fprintf(stderr, "   -s, --show        show uris\n");
        fprintf(stderr, "   -e, --erase       delete KB\n");
        fprintf(stderr, "   -f, --feature     enable features (default %s)\n", DEFAULT_FEATURES);
        fprintf(stderr, "       where features is one for more of\n");
	for (i=0; i<=TS_FEATURE_LAST; i++) {
	    fprintf(stderr, "          %c(%s)\n", *ts_feature_names[i], ts_feature_names[i]+1);
	}
        fprintf(stderr, "   -n, --print-only  dont execute commands, just show\n");
        fprintf(stderr, "This command creates new 3store users and/or KBs.\n");
	fprintf(stderr, "See man pages for more details\n");
        return 1;
    }

    /* Check if user exists and can access the KB */
    mdb = mysql_init(NULL);

    mysql_db = g_strdup_printf("%s%s", DBPREFIX, tsp_get_db());
    if (erase) {
	if (!mysql_real_connect(mdb, tsp_get_host(), tsp_get_user(),
			    tsp_get_pw(), mysql_db, 0, NULL, 0)) {
	    if (mysql_errno(mdb)/* != 1044 */) {
		fprintf(stderr, "Failed to connect to database: Error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		exit(1);
	    }
	} else {
	    erase_db(mdb);
	    LOG(1, "Database erased\n");
	}
    }

    while (!mysql_real_connect(mdb, tsp_get_host(), tsp_get_user(),
			    tsp_get_pw(), mysql_db, 0, NULL, 0)) {
	/* User cant access db, or something else is wrong */
	int err = mysql_errno(mdb);

	switch(err) {
	case CR_CONN_HOST_ERROR:
	case CR_CONNECTION_ERROR:
	    fprintf(stderr, "Could not connect to mysql server - ensure server is running\n");
	    exit(1);
	    break;
	case CR_UNKNOWN_HOST:
	    fprintf(stderr, "Unknown server host name.\n");
	    exit(1);
	    break;
	case CR_VERSION_ERROR:
	    fprintf(stderr, "Protocol mismatch - this verison of 3store was not compiled against the correct mysql version.\n");
	    exit(1);
	    break;
	case ER_DBACCESS_DENIED_ERROR:
	case ER_ACCESS_DENIED_ERROR:
	    /* These are expeted errors if the system has never been set up */
	    create_db(mdb);
	    create_user(mdb);
	    break;
	case ER_BAD_DB_ERROR:
	    /* This error is expected if the user exists but the database
             * doesnt, eg. its been dropped */
	    create_db(mdb);
	    create_user(mdb);
	    break;
	default:
	    fprintf(stderr, "Failed to connect to database: Error %d: %s\n", err, mysql_error(mdb));
	    exit(1);
	    break;
	}
    }

    if (!db_exists) create_tables(mdb, erase);

    fakecon.db = mdb;
    if (strchr(features, 'S') && strchr(features, 'I')) {
	char *std;

	fprintf(stderr, "Standards complaince mode and inference cannot be used at once\ndisabling standards compliance mode\n");
	std = strchr(features, 'S');
	*std = 'x';
    }
    if (!ts_config_get_value(&fakecon, "features", tmp) || features_set) {
	ts_config_set_value(&fakecon, "features", features);
    }

    con = tsp_connect(NULL);
    if (!db_exists && strchr(features, 'I')) {
	/* load definition of RDF */
	ts_erase(con, "http://www.w3.org/2000/01/rdf-schema#", TS_MODEL_DATA,
		 verbosity, stdout);
	ts_import(con, "file://" PKG_DATA_DIR "/rdf-schema.rdf",
		  "http://www.w3.org/2000/01/rdf-schema#", verbosity, 
		  stdout, TS_XML);
    }

    return 0;
}

void create_db(MYSQL *mdb)
{
    char *tmp;

    LOG(1, "Database not created, attempting to create.\n");
    if (!mysql_real_connect(mdb, tsp_get_host(), "root", "", "mysql",
			    0, NULL, 0)) {
	int i;
	char *pw;
	MYSQL *res = NULL;

	LOG(1, "MySQL root password required for 'mysql' database.\n");
	for (i=0; i<3 && !res; i++) {
	    pw = readline("MySQL root password: ");
	    res = mysql_real_connect(mdb, tsp_get_host(), "root", pw, "mysql",
			    0, NULL, 0);
	    /* due dilligence - dont really want the root pw kicking around */
	    memset(pw, 0, strlen(pw));
	    free(pw);
	}
	if (!res) {
	    fprintf(stderr, "Too many password attempts, aborting.\n");
	    exit(1);
	}
    }
    LOG(1, "Logged in as root\n");

    tmp = g_strdup_printf("CREATE DATABASE %s%s", DBPREFIX, tsp_get_db());
    if (QUERY(mdb, tmp)) {
	if (mysql_errno(mdb) == ER_DB_CREATE_EXISTS) {
	    LOG(1, "Database allready exists. Good.\n");
	    db_exists = 1;
	} else {
	    fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	    exit(1);
	}
    }
    free(tmp);
}

void erase_db(MYSQL *mdb)
{
    ts_connection c;
    char *tmp;

    if (!mysql_real_connect(mdb, tsp_get_host(), "root", "", "mysql",
			    0, NULL, 0)) {
	int i;
	char *pw;
	MYSQL *res = NULL;
	struct termios oldtty, newtty;

	LOG(1, "MySQL root password required for 'mysql' database.\n");
	for (i=0; i<3 && !res; i++) {
	    tcgetattr(fileno(stdout), &oldtty);
	    newtty = oldtty;
	    newtty.c_lflag &= ~ECHO;
	    tcsetattr(fileno(stdout), TCSANOW, &newtty);
	    pw = readline("MySQL root password: ");
	    tcsetattr(fileno(stdout), TCSANOW, &oldtty);
	    printf("\n");
	    res = mysql_real_connect(mdb, tsp_get_host(), "root", pw, "mysql",
			    0, NULL, 0);
	    /* due dilligence - dont really want the root pw kicking around */
	    memset(pw, 0, strlen(pw));
	    free(pw);
	}
	if (!res) {
	    fprintf(stderr, "Too many password attempts, aborting.\n");
	    exit(1);
	}
    }
    LOG(1, "Logged in as root\n");

    tmp = g_strdup_printf("DROP DATABASE %s%s", DBPREFIX, tsp_get_db());
    if (QUERY(mdb, tmp)) {
	fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	exit(1);
    }
    free(tmp);

    c.db = &mdb;

    ts_bulk_cache_flush(&c, 0, stdout);
}

void create_user(MYSQL *mdb)
{
    char *tmp;

    LOG(1, "User not created, or not allowed to access specified KB, attempting\nto create user.\n");
    tmp = g_strdup_printf("GRANT ALL PRIVILEGES ON %s%s.* TO %s@localhost IDENTIFIED BY '%s'", DBPREFIX, tsp_get_db(), tsp_get_user(), tsp_get_pw());
    if (QUERY(mdb, tmp)) {
	fprintf(stderr, "MySQL error %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
	exit(1);
    }
    free(tmp);
    LOG(1, "Created user and given permissions.\n");
}

void create_tables(MYSQL *mdb, int erase)
{
    FILE *tables;
    char tmp[4096], *c;
    int pos = 0;
    int table_version;
    ts_connection fake_c;

    LOG(1, "Creating tables.\n");
    snprintf(tmp, 4095, "USE %s%s", DBPREFIX, tsp_get_db());
    if (QUERY(mdb, tmp)) {
       fprintf(stderr, "MySQL error attempting to swtich to new db: %d: %s\n",
	       mysql_errno(mdb), mysql_error(mdb));
       exit(1);
    }

    if(!erase) {
	// we are in errase mode cannot refer to some tables
	table_version = ts_config_get_table_version(mdb);
    } 
    else {
	// create the current version for default
	table_version = TS_CURRENT_TABLE_VERSION;
    }
    
    if (TS_CURRENT_TABLE_VERSION > table_version && table_version > 0) {
	if (table_version < 2) {
	    LOG(1, "Upgrading tables to version 2");
	    QUERY(mdb, "DROP TABLE "TBLPREFIX"settings");

	    if (QUERY(mdb, "ALTER TABLE "TBLPREFIX"symbols "
			    "ADD INDEX(`bool`)")) {
		fprintf(stderr, "MySQL error upgrading symbols table %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		exit(1);
	    }
	}
	if (table_version < 3) {
	    LOG(1, "Upgrading tables to version 3");
	    if (QUERY(mdb, "ALTER TABLE "TBLPREFIX"triples "
			    "ADD COLUMN inf int(1) NOT NULL")) {
		fprintf(stderr, "MySQL error upgrading triples table %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		exit(1);
	    }
	    if (QUERY(mdb, "ALTER TABLE "TBLPREFIX"triples "
			    "ADD INDEX(inf)")) {
		fprintf(stderr, "MySQL error upgrading triples table %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		exit(1);
	    }
	}
	if (table_version < 6) {
	    LOG(1, "Upgrading tables to version 6");
	    if (QUERY(mdb, "ALTER TABLE "TBLPREFIX"symbols "
			    "ADD COLUMN ebv int(1) default NULL")) {
		fprintf(stderr, "MySQL error upgrading triples table %d: %s\n", mysql_errno(mdb), mysql_error(mdb));
		exit(1);
	    }
	}
    }

    if (TS_CURRENT_TABLE_VERSION < table_version && table_version > 0) {
	fprintf(stderr, "downgrading table versions is not supported, please"
		"create a new store if you wish to use an earlier version\n");
	exit(1);
    }

#ifdef ENABLE_QUADS
#define TABLE_FILE PKG_DATA_DIR "/tables-quad.sql"
#else
#define TABLE_FILE PKG_DATA_DIR "/tables.sql"
#endif
    if (!(tables = fopen(TABLE_FILE, "r"))) {
	perror("Cannot open table description file "
		TABLE_FILE);
	fprintf(stderr, "Cannot create KB tables.\n"
		        "Have you run 'make install'?\n");
	exit(1);
    }
    if (verbosity > 0) {
	printf("Reading table description from "TABLE_FILE"\n");
    }

    while (fgets(tmp+pos, sizeof(tmp) - pos, tables)) {
	char *p;
	int version;

	if (*tmp == '-') {
	    version = -1;
	    if (sscanf(tmp, "-- 3store: table-version %d", &version) > 0) {
		if (version < TS_CURRENT_TABLE_VERSION) {
		    fprintf(stderr, "Installed table descriptions are older than this software version.\nHave you run make install?\n");

		    exit(1);
		}
	    }

	    *tmp = '\0';
	}

	pos = strlen(tmp);
	if (strchr(tmp, ';')) {
	    for (c=tmp; *c; c++) {
		if (isspace(*c)) {
		    *c = ' ';
		}
		if (*c == ';') {
		    *c = '\0';
		}
	    }
	    if (MYSQL_VERSION_ID > 40100) {
		strncat(tmp, " DEFAULT CHARSET=utf8", 4096);
	    }

	    p = strstr(tmp, "$booltype$");
	    if (p) {
#if MYSQL_VERSION_ID >= 40100
		memcpy(p, "boolean   ", 10);
#else
		memcpy(p, "int(1)    ", 10);
#endif
	    }

	    LOG(2, "SQL: ");
	    LOG(2, tmp);
	    LOG(2, "\n");
	    if (QUERY(mdb, tmp)) {
		if (mysql_errno(mdb) == ER_TABLE_EXISTS_ERROR) {
		    LOG(1, "%s: looks like table is allready created.\n",
			   mysql_error(mdb));
		} else {
		    fprintf(stderr, "MySQL error in '%s' %d: %s\n", tmp, mysql_errno(mdb), mysql_error(mdb));
		    exit(1);
		}
	    }
	    pos = 0;
	    *tmp = '\0';
	}
    }

    if (TS_CURRENT_TABLE_VERSION > table_version && table_version > 1000) {
	if (table_version < 1005) {
	    printf("upgrading tables from to version 1005\n");
	    QUERY(mdb, "INSERT INTO "T_I" SELECT sid, `integer` FROM "T_S" WHERE `integer` IS NOT NULL");
	    QUERY(mdb, "INSERT INTO "T_F" SELECT sid, `floating` FROM "T_S" WHERE `floating` IS NOT NULL");
	    QUERY(mdb, "INSERT INTO "T_A" SELECT sid, `datetime` FROM "T_S" WHERE `datetime` IS NOT NULL");
	    QUERY(mdb, "ALTER TABLE "T_S" DROP COLUMN `integer`, DROP COLUMN `floating`, DROP COLUMN `datetime`, DROP COLUMN `bool`");
	}
    }

    ts_config_set_table_version(mdb, TS_CURRENT_TABLE_VERSION);
    fake_c.db = mdb;
    ts_config_set_value(&fake_c, "3store_version", PACKAGE_VERSION);
}

/* vi:set ts=8 sts=4 sw=4: */
